@@ -154,6 +154,8 @@ class RpcMethodEntry(NamedTuple):
 
 @runtime_checkable
 class RpcMethod(Protocol):
+    __slots__ = "__rpc_method__"
+
     __rpc_method__: RpcMethodEntry
 
 
@@ -264,7 +266,12 @@ class RpcRegistry:
                 )
                 for method, rpc_method in map(
                     lambda m1: (m1, cast(RpcMethod, m1)),
-                    filter(lambda m2: isinstance(m2, RpcMethod), iter_methods(obj)),
+                    iter_methods(
+                        obj,
+                        lambda m2: isinstance(m2, RpcMethod)
+                        or inspect.ismethod(m2)
+                        and isinstance(m2.__func__, RpcMethod),
+                    ),
                 )
             }
 
@@ -461,7 +468,7 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         try:
             b = body.decode(charset)
 
-            self._data_logger.trace(lambda: f"JSON Received: {b}")
+            self._data_logger.trace(lambda: f"JSON Received: {b!r}")
 
             self._handle_messages(self._generate_json_rpc_messages_from_dict(json.loads(b)))
         except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
@@ -528,7 +535,7 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         if self.write_transport is not None:
             msg = header + body
 
-            self._data_logger.trace(lambda: f"JSON send: {msg.decode()}")
+            self._data_logger.trace(lambda: f"JSON send: {msg.decode()!r}")
 
             if self._loop:
                 self._loop.call_soon_threadsafe(self.write_transport.write, msg)
@@ -624,7 +631,65 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
 
     @__logger.call
     async def handle_error(self, message: JsonRPCError) -> None:
-        raise JsonRPCErrorException(message.error.code, message.error.message, message.error.data)
+        if message.id is None:
+            error = "Invalid response. Response id is null."
+            self.__logger.warning(error)
+            raise JsonRPCErrorException(message.error.code, message.error.message, message.error.data)
+
+        with self._sended_request_lock:
+            entry = self._sended_request.pop(message.id, None)
+
+        if entry is None:
+            error = f"Invalid response. Could not find id '{message.id}' in request list."
+            self.__logger.warning(error)
+            raise JsonRPCErrorException(message.error.code, message.error.message, message.error.data)
+
+        try:
+            if not entry.future.done():
+                res = None
+                if message.result is not None:
+                    res = from_dict(message.result, entry.result_type)
+                if entry.future.get_loop() == asyncio.get_running_loop():
+                    entry.future.set_exception(
+                        JsonRPCErrorException(message.error.code, message.error.message, message.error.data)
+                    )
+                else:
+                    if entry.future.get_loop().is_running():
+
+                        def set_result(f: asyncio.Future[Any], r: Any, ev: threading.Event) -> None:
+                            try:
+                                if not f.done() and f.get_loop().is_running():
+                                    f.set_exception(
+                                        JsonRPCErrorException(
+                                            message.error.code, message.error.message, message.error.data
+                                        )
+                                    )
+                            finally:
+                                ev.set()
+
+                        done = threading.Event()
+
+                        entry.future.get_loop().call_soon_threadsafe(set_result, entry.future, res, done)
+
+                        start = time.monotonic()
+                        while not done.is_set():
+                            if time.monotonic() - start > 120:
+                                raise TimeoutError("Can't set future result.")
+
+                            await asyncio.sleep(0)
+
+                    else:
+                        self.__logger.warning(lambda: f"Response {entry!r} loop is not running.")
+
+        except (SystemExit, KeyboardInterrupt):
+            raise
+        except BaseException as e:
+            if not entry.future.done():
+                if entry.future.get_loop() == asyncio.get_running_loop():
+                    entry.future.set_exception(e)
+                else:
+                    if entry.future.get_loop().is_running():
+                        entry.future.get_loop().call_soon_threadsafe(entry.future.set_exception, e)
 
     @staticmethod
     def _convert_params(
@@ -703,7 +768,13 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
 
         params = self._convert_params(e.method, e.param_type, message.params)
 
-        if isinstance(e.method, HasThreaded) and e.method.__threaded__:
+        if (
+            isinstance(e.method, HasThreaded)
+            and e.method.__threaded__
+            or inspect.ismethod(e.method)
+            and isinstance(e.method.__func__, HasThreaded)
+            and e.method.__func__.__threaded__
+        ):
             task = run_coroutine_in_thread(
                 ensure_coroutine(cast(Callable[..., Any], e.method)), *params[0], **params[1]
             )
@@ -768,7 +839,13 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         try:
             params = self._convert_params(e.method, e.param_type, message.params)
 
-            if isinstance(e.method, HasThreaded) and e.method.__threaded__:
+            if (
+                isinstance(e.method, HasThreaded)
+                and e.method.__threaded__
+                or inspect.ismethod(e.method)
+                and isinstance(e.method.__func__, HasThreaded)
+                and e.method.__func__.__threaded__
+            ):
                 task = run_coroutine_in_thread(
                     ensure_coroutine(cast(Callable[..., Any], e.method)), *params[0], **params[1]
                 )
